@@ -57,7 +57,7 @@ A commercial use license is available from Genivia, Inc., contact@genivia.com
 
 /**
 
-@page wsa_0 The WS-Addressing plugin
+@page wsa_0 The WS-Addressing Plugin
 
 [TOC]
 
@@ -288,7 +288,8 @@ additional parameter (SOAP Fault detail):
 WS-Security can be combined with WS-Addressing. To sign WS-Addressing header
 blocks, use the `soap_wsse_set_wsu_id` WSSE-plugin call to set the wsu:Id
 attribute and signing of these attributed elements. For example, suppose we use
-WS-Addressing 2005:
+WS-Addressing 2005 headers (which are activated with an `#import "wsa5.h"` in
+the header file for soapcpp2):
 
 @code
     #include "wsaapi.h"
@@ -296,7 +297,7 @@ WS-Addressing 2005:
     soap_register_plugin(soap, soap_wsa);
     soap_register_plugin(soap, soap_wsse);
 
-    soap_wsse_set_wsu_id(soap, "wsa5:From wsa5:To wsa5:ReplyTo wsa5:FaultTo wsa5:Action");
+    soap_wsse_set_wsu_id(soap, "wsa5:From wsa5:To wsa5:ReplyTo wsa5:FaultTo wsa5:Action wsa5:MessageID");
     if (soap_wsa_request(soap, RequestMessageID, ToAddress, RequestAction)
      || soap_wsa_add_From(soap, FromAddress) // optional: add a 'From' address
      || soap_wsa_add_FaultTo(soap, FaultToAddress))
@@ -305,6 +306,13 @@ WS-Addressing 2005:
       ... // error
 @endcode
 
+If your are using WS-Addressing 2004 (which is activated with an
+`#import "wsa.h"` in the header file for soapcpp2) then change one line:
+
+@code
+    soap_wsse_set_wsu_id(soap, "wsa:From wsa:To wsa:ReplyTo wsa:FaultTo wsa:Action wsa:MessageID");
+@endcode
+
 Note: `soap_wsse_set_wsu_id` should only be set once. Each new call overrides
 the previous.
 
@@ -577,7 +585,7 @@ static int soap_wsa_alloc_header(struct soap *soap);
 /**
 @fn const char *soap_wsa_rand_uuid(struct soap *soap)
 @brief Generates a random UUID (UUID algorithm version 4). Compile all source
-codes with -DWITH_OPENSSL for better randomness results.
+codes with -DWITH_OPENSSL for better randomness.
 @param soap context
 @return UUID "urn:uuid:xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx" or NULL if out of memory
 */
@@ -586,40 +594,7 @@ const char*
 SOAP_FMAC2
 soap_wsa_rand_uuid(struct soap *soap)
 {
-  const int uuidlen = 48;
-  char *uuid = (char*)soap_malloc(soap, uuidlen);
-  int r1, r2, r3, r4;
-#ifdef WITH_OPENSSL
-  r1 = soap_random;
-  r2 = soap_random;
-#else
-  size_t i;
-  static int k = 0xFACEB00B;
-  int lo = k % 127773;
-  int hi = k / 127773;
-# ifdef HAVE_GETTIMEOFDAY
-  struct timeval tv;
-  gettimeofday(&tv, NULL);
-  r1 = 10000000 * tv.tv_sec + tv.tv_usec;
-#else
-  r1 = (int)time(NULL);
-# endif
-  k = 16807 * lo - 2836 * hi;
-  if (k <= 0)
-    k += 0x7FFFFFFF;
-  r2 = k;
-  k &= 0x8FFFFFFF;
-  for (i = 0; i < (sizeof(soap->buf) < 16UL ? sizeof(soap->buf) : 16UL); i++)
-    r2 += soap->buf[i];
-#endif
-  r3 = soap_random;
-  r4 = soap_random;
-  if (uuid)
-  {
-    (SOAP_SNPRINTF(uuid, uuidlen, 45), "urn:uuid:%8.8x-%4.4hx-4%3.3hx-%4.4hx-%4.4hx%8.8x", r1, (short)(r2 >> 16), (short)(((short)r2 >> 4) & 0x0FFF), (short)(((short)(r3 >> 16) & 0x3FFF) | 0x8000), (short)r3, r4);
-    DBGFUN1("soap_wsa_rand_uuid", "%s", uuid);
-  }
-  return uuid;
+  return soap_strdup(soap, soap_rand_uuid(soap, "urn:uuid:"));
 }
 
 /******************************************************************************\
@@ -984,7 +959,7 @@ soap_wsa_reply(struct soap *soap, const char *id, const char *action)
       {
         soap_copy_stream(reply_soap, soap);
         soap_free_stream(soap); /* prevents close in soap_connect() below */
-        soap->omode |= SOAP_ENC_XML; /* omit HTTP header ("encode XML body only") */
+        soap->omode |= SOAP_ENC_PLAIN; /* omit HTTP header ("encode XML body only") */
         if (soap_connect(soap, newheader->SOAP_WSA(To), newheader->SOAP_WSA(Action)))
         {
           int err;
@@ -1006,7 +981,7 @@ soap_wsa_reply(struct soap *soap, const char *id, const char *action)
         if (soap_valid_socket(reply_soap->socket))
           soap_send_empty_response(reply_soap, SOAP_OK);        /* HTTP ACCEPTED */
         soap->header = newheader;
-        soap->omode &= ~SOAP_ENC_XML;           /* HTTP header required */
+        soap->omode &= ~SOAP_ENC_PLAIN;           /* HTTP header required */
         soap_end(reply_soap);
         soap_free(reply_soap);
         data->fresponse = soap->fresponse;
@@ -1065,11 +1040,8 @@ int
 SOAP_FMAC2
 soap_wsa_fault_subcode_action(struct soap *soap, int flag, const char *faultsubcode, const char *faultstring, const char *faultdetail, const char *action)
 {
-  struct soap_wsa_data *data = (struct soap_wsa_data*)soap_lookup_plugin(soap, soap_wsa_id);
   struct SOAP_ENV__Header *oldheader, *newheader;
-  DBGFUN2("soap_wsa_fault_subcode", "faultsubcode=%s", faultsubcode?faultsubcode:"(null)", "faultstring=%s", faultstring?faultstring:"(null)");
-  if (!data)
-    return soap->error = SOAP_PLUGIN_ERROR;
+  DBGFUN2("soap_wsa_fault_subcode_action", "faultsubcode=%s", faultsubcode ? faultsubcode : "(null)", "faultstring=%s", faultstring ? faultstring : "(null)");
   oldheader = soap->header;
   /* no FaultTo: use ReplyTo */
   if (oldheader && oldheader->SOAP_WSA(ReplyTo) && (!oldheader->SOAP_WSA(FaultTo) || soap_tagsearch(soap_wsa_allAnonymousURI, oldheader->SOAP_WSA(FaultTo)->Address)))
@@ -1083,51 +1055,53 @@ soap_wsa_fault_subcode_action(struct soap *soap, int flag, const char *faultsubc
     if (oldheader->SOAP_WSA(FaultTo))
       oldheader->SOAP_WSA(FaultTo)->Address = oldheader->SOAP_WSA(ReplyTo)->Address;
   }
-  if (oldheader && oldheader->SOAP_WSA(FaultTo))
-  {
-    DBGLOG(TEST, SOAP_MESSAGE(fdebug, "WSA FaultTo='%s'\n", oldheader->SOAP_WSA(FaultTo)->Address));
-  }
+  /* use FaultTo */
   if (oldheader && oldheader->SOAP_WSA(FaultTo) && !strcmp(oldheader->SOAP_WSA(FaultTo)->Address, soap_wsa_noneURI))
     return soap_send_empty_response(soap, SOAP_OK);     /* HTTP ACCEPTED */
   soap->header = NULL;
   /* allocate a new header */
-  if (soap_wsa_alloc_header(soap))
-    return soap->error;
+  soap_header(soap);
   newheader = soap->header;
-  soap_default_SOAP_ENV__Header(soap, newheader); /* remove/clear SOAP Header */
-  /* check header */
-  if (oldheader && oldheader->SOAP_WSA(MessageID))
+  if (newheader)
   {
-    newheader->SOAP_WSA(RelatesTo) = (SOAP_WSA_(,RelatesTo)*)soap_malloc(soap, sizeof(SOAP_WSA_(,RelatesTo)));
-    SOAP_WSA_(soap_default_,RelatesTo)(soap, newheader->SOAP_WSA(RelatesTo));
-    newheader->SOAP_WSA(RelatesTo)->__item = oldheader->SOAP_WSA(MessageID);
-  }
-  /* header->wsa__MessageID = "..."; */
-  newheader->SOAP_WSA(Action) = (char*)soap_wsa_faultAction;
-  if (oldheader && oldheader->SOAP_WSA(FaultTo) && oldheader->SOAP_WSA(FaultTo)->Address && !soap_tagsearch(soap_wsa_allAnonymousURI, oldheader->SOAP_WSA(FaultTo)->Address))
-  {
-    newheader->SOAP_WSA(To) = oldheader->SOAP_WSA(FaultTo)->Address;
-    /* (re)connect to FaultTo endpoint if From != FaultTo */
-    if (!oldheader->SOAP_WSA(From) || !oldheader->SOAP_WSA(From)->Address || strcmp(oldheader->SOAP_WSA(From)->Address, oldheader->SOAP_WSA(FaultTo)->Address))
+    soap_default_SOAP_ENV__Header(soap, newheader); /* remove/clear SOAP Header */
+    /* check header */
+    if (oldheader && oldheader->SOAP_WSA(MessageID))
     {
-      soap->keep_alive = 0;
-      soap_send_empty_response(soap, SOAP_OK);  /* HTTP ACCEPTED */
-      if (soap_connect(soap, newheader->SOAP_WSA(To), newheader->SOAP_WSA(Action)))
-        return soap->error = SOAP_STOP; /* nowhere to go */
-      soap_set_endpoint(soap, newheader->SOAP_WSA(To));
-      if (action)
-        soap->action = (char*)action;
-      else
-        soap->action = newheader->SOAP_WSA(Action);
-      data->fresponse = soap->fresponse;
-      soap->fresponse = soap_wsa_response;      /* response will be a POST */
+      newheader->SOAP_WSA(RelatesTo) = (SOAP_WSA_(,RelatesTo)*)soap_malloc(soap, sizeof(SOAP_WSA_(,RelatesTo)));
+      SOAP_WSA_(soap_default_,RelatesTo)(soap, newheader->SOAP_WSA(RelatesTo));
+      newheader->SOAP_WSA(RelatesTo)->__item = oldheader->SOAP_WSA(MessageID);
     }
+    /* header->wsa__MessageID = "..."; */
+    newheader->SOAP_WSA(Action) = (char*)soap_wsa_faultAction;
+    if (oldheader && oldheader->SOAP_WSA(FaultTo) && oldheader->SOAP_WSA(FaultTo)->Address && !soap_tagsearch(soap_wsa_allAnonymousURI, oldheader->SOAP_WSA(FaultTo)->Address))
+    {
+      newheader->SOAP_WSA(To) = oldheader->SOAP_WSA(FaultTo)->Address;
+      /* (re)connect to FaultTo endpoint if From != FaultTo */
+      if (!oldheader->SOAP_WSA(From) || !oldheader->SOAP_WSA(From)->Address || strcmp(oldheader->SOAP_WSA(From)->Address, oldheader->SOAP_WSA(FaultTo)->Address))
+      {
+	struct soap_wsa_data *data = (struct soap_wsa_data*)soap_lookup_plugin(soap, soap_wsa_id);
+	if (!data)
+	  return soap->error = SOAP_PLUGIN_ERROR;
+	soap->keep_alive = 0;
+	soap_send_empty_response(soap, SOAP_OK);  /* HTTP ACCEPTED */
+	if (soap_connect(soap, newheader->SOAP_WSA(To), newheader->SOAP_WSA(Action)))
+	  return soap->error = SOAP_STOP; /* nowhere to go */
+	soap_set_endpoint(soap, newheader->SOAP_WSA(To));
+	if (action)
+	  soap->action = (char*)action;
+	else
+	  soap->action = newheader->SOAP_WSA(Action);
+	data->fresponse = soap->fresponse;
+	soap->fresponse = soap_wsa_response;      /* response will be a POST */
+      }
+    }
+    else if (oldheader && oldheader->SOAP_WSA(From))
+      newheader->SOAP_WSA(To) = oldheader->SOAP_WSA(From)->Address;
+    else
+      newheader->SOAP_WSA(To) = (char*)soap_wsa_anonymousURI;
+    soap->header = newheader;
   }
-  else if (oldheader && oldheader->SOAP_WSA(From))
-    newheader->SOAP_WSA(To) = oldheader->SOAP_WSA(From)->Address;
-  else
-    newheader->SOAP_WSA(To) = (char*)soap_wsa_anonymousURI;
-  soap->header = newheader;
   if (flag)
     return soap_sender_fault_subcode(soap, faultsubcode, faultstring, faultdetail);
   return soap_receiver_fault_subcode(soap, faultsubcode, faultstring, faultdetail);
@@ -1382,7 +1356,7 @@ soap_wsa_error(struct soap *soap, SOAP_WSA(FaultCodesType) fault, const char *in
         soap->fault->detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemHeaderQName);
         soap->fault->detail->fault = (void*)info;
       }
-      else
+      else if (soap->version == 2)
       {
         soap->fault->SOAP_ENV__Detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemHeaderQName);
         soap->fault->SOAP_ENV__Detail->fault = (void*)info;
@@ -1414,7 +1388,7 @@ soap_wsa_error(struct soap *soap, SOAP_WSA(FaultCodesType) fault, const char *in
         soap->fault->detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemHeaderQName);
         soap->fault->detail->fault = (void*)info;
       }
-      else
+      else if (soap->version == 2)
       {
         soap->fault->SOAP_ENV__Detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemHeaderQName);
         soap->fault->SOAP_ENV__Detail->fault = (void*)info;
@@ -1428,7 +1402,7 @@ soap_wsa_error(struct soap *soap, SOAP_WSA(FaultCodesType) fault, const char *in
         soap->fault->detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemIRI);
         soap->fault->detail->fault = (void*)info;
       }
-      else
+      else if (soap->version == 2)
       {
         soap->fault->SOAP_ENV__Detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemIRI);
         soap->fault->SOAP_ENV__Detail->fault = (void*)info;
@@ -1444,7 +1418,7 @@ soap_wsa_error(struct soap *soap, SOAP_WSA(FaultCodesType) fault, const char *in
         SOAP_WSA_(soap_default_,ProblemAction)(soap, (SOAP_WSA_(,ProblemAction)*)soap->fault->detail->fault);
         ((SOAP_WSA_(,ProblemAction)*)soap->fault->detail->fault)->Action = (char*)info;
       }
-      else
+      else if (soap->version == 2)
       {
         soap->fault->SOAP_ENV__Detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemAction);
         soap->fault->SOAP_ENV__Detail->fault = (void*)soap_malloc(soap, sizeof(SOAP_WSA_(,ProblemAction)));
@@ -1460,7 +1434,7 @@ soap_wsa_error(struct soap *soap, SOAP_WSA(FaultCodesType) fault, const char *in
         soap->fault->detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemIRI);
         soap->fault->detail->fault = (void*)info;
       }
-      else
+      else if (soap->version == 2)
       {
         soap->fault->SOAP_ENV__Detail->__type = SOAP_WSA_(SOAP_TYPE_,ProblemIRI);
         soap->fault->SOAP_ENV__Detail->fault = (void*)info;
@@ -1639,7 +1613,7 @@ static void
 soap_wsa_set_error(struct soap *soap, const char **c, const char **s)
 {
   struct soap_wsa_data *data = (struct soap_wsa_data*)soap_lookup_plugin(soap, soap_wsa_id);
-  DBGFUN("soap_wsa_set_error");
+  DBGFUN2("soap_wsa_set_error", "code=%s", c && *c ? *c : "(null)", "string=%s", s && *s ? *s : "(null)");
   if (!data)
     return;
   if (data->fseterror)
